@@ -1,6 +1,11 @@
 //! Tokenization utilities for building parsers in Rust
 
-use std::{collections::VecDeque, sync::mpsc::{sync_channel, Receiver, RecvError, SyncSender}, usize};
+use std::{collections::VecDeque, fmt::Debug, usize};
+
+pub use buffered_token_queue::*;
+pub use generator_token_queue::*;
+#[cfg(not(target_arch = "wasm32"))]
+pub use parallel_token_queue::*;
 
 /// A structure with a piece of data and some additional data such as a position
 pub struct Token<T: PartialEq, TData>(pub T, pub TData);
@@ -43,177 +48,270 @@ pub trait TokenSender<T: PartialEq, TData> {
     fn push(&mut self, token: Token<T, TData>);
 }
 
-/// A queue which can be used as a sender and reader. Use this for buffering all the tokens before reading
-pub struct BufferedTokenQueue<T: PartialEq, TData> {
-    buffer: VecDeque<Token<T, TData>>,
-}
-
-impl<T: PartialEq, TData> BufferedTokenQueue<T, TData> {
-    /// Constructs a new [`BufferedTokenQueue`]
-    pub fn new() -> Self {
-        BufferedTokenQueue {
-            buffer: VecDeque::new(),
-        }
-    }
-}
-
-impl<T: PartialEq, TData> TokenSender<T, TData> for BufferedTokenQueue<T, TData> {
-    fn push(&mut self, token: Token<T, TData>) {
-        self.buffer.push_back(token)
-    }
-}
-
-impl<T: PartialEq, TData> TokenReader<T, TData> for BufferedTokenQueue<T, TData> {
-    fn peek(&mut self) -> Option<&Token<T, TData>> {
-        self.buffer.front()
+mod buffered_token_queue {
+    use super::*;
+    /// A queue which can be used as a sender and reader. Use this for buffering all the tokens before reading
+    pub struct BufferedTokenQueue<T: PartialEq, TData> {
+        buffer: VecDeque<Token<T, TData>>,
     }
 
-    fn next(&mut self) -> Option<Token<T, TData>> {
-        self.buffer.pop_front()
-    }
-
-    fn scan(&mut self, mut cb: impl FnMut(&T, &TData) -> bool) -> Option<&Token<T, TData>> {
-        let mut iter = self.buffer.iter().peekable();
-        while let Some(token) = iter.next() {
-            if cb(&token.0, &token.1) {
-                return iter.peek().map(|v| *v);
+    impl<T: PartialEq, TData> BufferedTokenQueue<T, TData> {
+        /// Constructs a new [`BufferedTokenQueue`]
+        pub fn new() -> Self {
+            BufferedTokenQueue {
+                buffer: VecDeque::new(),
             }
         }
-        None
+    }
+
+    impl<T: PartialEq, TData> TokenSender<T, TData> for BufferedTokenQueue<T, TData> {
+        fn push(&mut self, token: Token<T, TData>) {
+            self.buffer.push_back(token)
+        }
+    }
+
+    impl<T: PartialEq, TData> TokenReader<T, TData> for BufferedTokenQueue<T, TData> {
+        fn peek(&mut self) -> Option<&Token<T, TData>> {
+            self.buffer.front()
+        }
+
+        fn next(&mut self) -> Option<Token<T, TData>> {
+            self.buffer.pop_front()
+        }
+
+        fn scan(&mut self, mut cb: impl FnMut(&T, &TData) -> bool) -> Option<&Token<T, TData>> {
+            let mut iter = self.buffer.iter().peekable();
+            while let Some(token) = iter.next() {
+                if cb(&token.0, &token.1) {
+                    return iter.peek().map(|v| *v);
+                }
+            }
+            None
+        }
     }
 }
 
-const STREAMED_CHANNEL_BUFFER_SIZE: usize = 20;
+#[cfg(not(target_arch = "wasm32"))]
+mod parallel_token_queue {
+    use super::*;
+    use std::sync::mpsc::{sync_channel, Receiver, RecvError, SyncSender};
 
-/// A token queue used for doing lexing and parsing on different threads. Will send tokens between threads
-pub struct ParallelTokenQueue;
+    const STREAMED_CHANNEL_BUFFER_SIZE: usize = 20;
 
-impl ParallelTokenQueue {
-    /// Creates two items, a sender and a receiver. Where the reader is on the parsing thread and the 
-    /// sender is on the lexer thread
-    pub fn new<T: PartialEq, TData>(
-    ) -> (ParallelTokenSender<T, TData>, ParallelTokenReader<T, TData>) {
-        let (sender, receiver) = sync_channel::<Token<T, TData>>(STREAMED_CHANNEL_BUFFER_SIZE);
-        (
-            ParallelTokenSender(sender),
-            ParallelTokenReader {
-                receiver,
-                cache: VecDeque::new(),
-            },
-        )
+    /// A token queue used for doing lexing and parsing on different threads. Will send tokens between threads
+    pub struct ParallelTokenQueue;
+
+    impl ParallelTokenQueue {
+        /// Creates two items, a sender and a receiver. Where the reader is on the parsing thread and the
+        /// sender is on the lexer thread
+        pub fn new<T: PartialEq, TData>(
+        ) -> (ParallelTokenSender<T, TData>, ParallelTokenReader<T, TData>) {
+            let (sender, receiver) = sync_channel::<Token<T, TData>>(STREAMED_CHANNEL_BUFFER_SIZE);
+            (
+                ParallelTokenSender(sender),
+                ParallelTokenReader {
+                    receiver,
+                    cache: VecDeque::new(),
+                },
+            )
+        }
     }
-}
 
-// Sender and reader structs generate by `ParallelTokenQueue::new`:
+    // Sender and reader structs generate by `ParallelTokenQueue::new`:
 
-#[doc(hidden)]
-pub struct ParallelTokenSender<T: PartialEq, TData>(SyncSender<Token<T, TData>>);
+    #[doc(hidden)]
+    pub struct ParallelTokenSender<T: PartialEq, TData>(SyncSender<Token<T, TData>>);
 
-#[doc(hidden)]
-pub struct ParallelTokenReader<T: PartialEq, TData> {
-    receiver: Receiver<Token<T, TData>>,
-    cache: VecDeque<Token<T, TData>>,
-}
-
-impl<T: PartialEq, TData> TokenSender<T, TData> for ParallelTokenSender<T, TData> {
-    fn push(&mut self, token: Token<T, TData>) {
-        self.0.send(token).unwrap();
+    #[doc(hidden)]
+    pub struct ParallelTokenReader<T: PartialEq, TData> {
+        receiver: Receiver<Token<T, TData>>,
+        cache: VecDeque<Token<T, TData>>,
     }
-}
 
-impl<T: PartialEq, TData> TokenReader<T, TData> for ParallelTokenReader<T, TData> {
-    fn peek(&mut self) -> Option<&Token<T, TData>> {
-        if self.cache.is_empty() {
-            match self.receiver.recv() {
-                Ok(token) => self.cache.push_back(token),
-                // Err is reader has dropped e.g. no more tokens
-                Err(RecvError) => {
-                    return None;
+    impl<T: PartialEq, TData> TokenSender<T, TData> for ParallelTokenSender<T, TData> {
+        fn push(&mut self, token: Token<T, TData>) {
+            self.0.send(token).unwrap();
+        }
+    }
+
+    impl<T: PartialEq, TData> TokenReader<T, TData> for ParallelTokenReader<T, TData> {
+        fn peek(&mut self) -> Option<&Token<T, TData>> {
+            if self.cache.is_empty() {
+                match self.receiver.recv() {
+                    Ok(token) => self.cache.push_back(token),
+                    // Err is reader has dropped e.g. no more tokens
+                    Err(RecvError) => {
+                        return None;
+                    }
+                }
+            }
+            self.cache.front()
+        }
+
+        fn next(&mut self) -> Option<Token<T, TData>> {
+            if !self.cache.is_empty() {
+                return self.cache.pop_front();
+            }
+            self.receiver.recv().ok()
+        }
+
+        fn scan(&mut self, mut cb: impl FnMut(&T, &TData) -> bool) -> Option<&Token<T, TData>> {
+            let found = scan_cache(&mut self.cache, &mut cb);
+            let mut return_next = match found {
+                ScanCacheResult::FoundInCache(i) => return self.cache.get(i),
+                ScanCacheResult::NotFound => false,
+                ScanCacheResult::Found => true,
+            };
+            loop {
+                match self.receiver.recv() {
+                    Ok(val) => {
+                        if return_next {
+                            self.cache.push_back(val);
+                            return self.cache.back();
+                        }
+                        if (&mut cb)(&val.0, &val.1) {
+                            return_next = true;
+                        }
+                        self.cache.push_back(val);
+                    }
+                    // Err is reader has dropped e.g. no more tokens
+                    Err(RecvError) => {
+                        return None;
+                    }
                 }
             }
         }
-        self.cache.front()
-    }
-
-    fn next(&mut self) -> Option<Token<T, TData>> {
-        // SAFETY: safe to get mutable reference for this function as have mutable self
-        // let cache = unsafe { &mut *self.cache.get() };
-        if !self.cache.is_empty() {
-            return self.cache.pop_front();
-        }
-        self.receiver.recv().ok()
-    }
-
-    fn scan(&mut self, cb: impl FnMut(&T, &TData) -> bool) -> Option<&Token<T, TData>> {
-        let reciever = &mut self.receiver;
-        scan(&mut self.cache, cb, || reciever.recv().ok())
     }
 }
 
-/// A token queue which has a backing generator/lexer which is called when needed by parsing logic 
-pub struct GeneratorTokenQueue<
-    T: PartialEq,
-    TData,
-    TGeneratorState,
-    TGenerator: FnMut(&mut TGeneratorState) -> Option<Token<T, TData>>,
-> {
-    generator: TGenerator,
-    generator_state: TGeneratorState,
-    cache: VecDeque<Token<T, TData>>,
-}
+mod generator_token_queue {
+    use super::*;
 
-impl<
+    /// A token queue which has a backing generator/lexer which is called when needed by parsing logic
+    pub struct GeneratorTokenQueue<T, TData, TGeneratorState, TGenerator>
+    where
         T: PartialEq,
-        TData,
-        TGeneratorState,
-        TGenerator: FnMut(&mut TGeneratorState) -> Option<Token<T, TData>>,
-    > GeneratorTokenQueue<T, TData, TGeneratorState, TGenerator>
-{
-    /// Create a new [`GeneratorTokenQueue`] with a lexer function and initial state
-    pub fn new(generator: TGenerator, generator_state: TGeneratorState) -> Self {
-        GeneratorTokenQueue {
-            generator,
-            generator_state,
-            cache: VecDeque::new(),
-        }
-    }
-}
-
-impl<
-        T: PartialEq + std::fmt::Debug,
-        TData,
-        TGeneratorState: std::fmt::Debug,
-        TGenerator: FnMut(&mut TGeneratorState) -> Option<Token<T, TData>>,
-    > TokenReader<T, TData> for GeneratorTokenQueue<T, TData, TGeneratorState, TGenerator>
-{
-    fn next(&mut self) -> Option<Token<T, TData>> {
-        if !self.cache.is_empty() {
-            return self.cache.pop_front();
-        }
-        (self.generator)(&mut self.generator_state)
+        for<'a> TGenerator:
+            FnMut(&mut TGeneratorState, &mut GeneratorTokenQueueBuffer<'a, T, TData>) -> (),
+    {
+        generator: TGenerator,
+        generator_state: TGeneratorState,
+        cache: VecDeque<Token<T, TData>>,
     }
 
-    fn peek(&mut self) -> Option<&Token<T, TData>> {
-        if self.cache.is_empty() {
-            if let Some(token) = (self.generator)(&mut self.generator_state) {
-                self.cache.push_back(token)
+    /// A wrapping struct for the cache around [`GeneratorTokenQueue`]. Use as the second parameter
+    /// in the generator/lexer function
+    pub struct GeneratorTokenQueueBuffer<'a, T: PartialEq, TData>(
+        &'a mut VecDeque<Token<T, TData>>,
+    );
+
+    impl<'a, T: PartialEq, TData> GeneratorTokenQueueBuffer<'a, T, TData> {
+        pub fn is_empty(&self) -> bool {
+            self.0.is_empty()
+        }
+
+        pub fn len(&self) -> usize {
+            self.0.len()
+        }
+    }
+
+    impl<'a, T: PartialEq, TData> TokenSender<T, TData> for GeneratorTokenQueueBuffer<'a, T, TData> {
+        fn push(&mut self, token: Token<T, TData>) {
+            self.0.push_back(token);
+        }
+    }
+
+    impl<T, TData, TGeneratorState, TGenerator>
+        GeneratorTokenQueue<T, TData, TGeneratorState, TGenerator>
+    where
+        T: PartialEq,
+        for<'a> TGenerator:
+            FnMut(&mut TGeneratorState, &mut GeneratorTokenQueueBuffer<'a, T, TData>) -> (),
+    {
+        /// Create a new [`GeneratorTokenQueue`] with a lexer function and initial state
+        pub fn new(generator: TGenerator, generator_state: TGeneratorState) -> Self {
+            GeneratorTokenQueue {
+                generator,
+                generator_state,
+                cache: VecDeque::new(),
             }
         }
-        self.cache.front()
     }
 
-    fn scan(&mut self, cb: impl FnMut(&T, &TData) -> bool) -> Option<&Token<T, TData>> {
-        let generator = &mut self.generator;
-        let state = &mut self.generator_state;
-        scan(&mut self.cache, cb, || (generator)(state))
+    impl<T, TData, TGeneratorState, TGenerator> TokenReader<T, TData>
+        for GeneratorTokenQueue<T, TData, TGeneratorState, TGenerator>
+    where
+        T: PartialEq + Debug,
+        TData: Debug,
+        for<'a> TGenerator:
+            FnMut(&mut TGeneratorState, &mut GeneratorTokenQueueBuffer<'a, T, TData>) -> (),
+    {
+        fn next(&mut self) -> Option<Token<T, TData>> {
+            if !self.cache.is_empty() {
+                return self.cache.pop_front();
+            }
+            (self.generator)(
+                &mut self.generator_state,
+                &mut GeneratorTokenQueueBuffer(&mut self.cache),
+            );
+            self.cache.pop_front()
+        }
+
+        fn peek(&mut self) -> Option<&Token<T, TData>> {
+            if self.cache.is_empty() {
+                (self.generator)(
+                    &mut self.generator_state,
+                    &mut GeneratorTokenQueueBuffer(&mut self.cache),
+                );
+            }
+            self.cache.front()
+        }
+
+        fn scan(&mut self, mut cb: impl FnMut(&T, &TData) -> bool) -> Option<&Token<T, TData>> {
+            let cb = &mut cb;
+            let found = scan_cache(&mut self.cache, cb);
+            let mut return_next = match found {
+                ScanCacheResult::FoundInCache(i) => return self.cache.get(i),
+                ScanCacheResult::NotFound => false,
+                ScanCacheResult::Found => true,
+            };
+            let mut found = None::<usize>;
+            while found.is_none() {
+                (self.generator)(
+                    &mut self.generator_state,
+                    &mut GeneratorTokenQueueBuffer(&mut self.cache),
+                );
+                match self.cache.back() {
+                    Some(token) => {
+                        if return_next {
+                            found = Some(self.cache.len() - 1);
+                            break;
+                        }
+                        if cb(&token.0, &token.1) {
+                            return_next = true;
+                        }
+                    }
+                    None => {
+                        return None;
+                    }
+                }
+            }
+            self.cache.get(found.unwrap())
+        }
     }
 }
 
-fn scan<T: PartialEq, TData>(
+enum ScanCacheResult<T> {
+    NotFound,
+    FoundInCache(T),
+    // Aka pull out next one
+    Found,
+}
+
+fn scan_cache<T: PartialEq, TData>(
     cache: &mut VecDeque<Token<T, TData>>,
-    mut cb: impl FnMut(&T, &TData) -> bool,
-    mut generator: impl FnMut() -> Option<Token<T, TData>>,
-) -> Option<&Token<T, TData>> {
+    cb: &mut impl FnMut(&T, &TData) -> bool,
+) -> ScanCacheResult<usize> {
     // Scan cache first
     let mut found = None;
     for (i, token) in cache.iter().enumerate() {
@@ -224,27 +322,13 @@ fn scan<T: PartialEq, TData>(
     }
     if let Some(i) = found {
         if i < cache.len() {
-            return cache.get(i);
+            return ScanCacheResult::FoundInCache(i);
         }
     }
-    let mut found = found.is_some();
-    // Generate and add to cache while not found
-    loop {
-        match (generator)() {
-            Some(token) => {
-                if found {
-                    cache.push_back(token);
-                    return cache.back();
-                }
-                if cb(&token.0, &token.1) {
-                    found = true;
-                }
-                cache.push_back(token);
-            }
-            None => {
-                return None;
-            }
-        }
+    if found.is_some() {
+        ScanCacheResult::Found
+    } else {
+        ScanCacheResult::NotFound
     }
 }
 
@@ -270,7 +354,7 @@ mod tests {
 
     impl<T: PartialEq, TData: PartialEq> Eq for Token<T, TData> {}
 
-    mod static_token_channel {
+    mod buffered_token_queue {
         use super::{BufferedTokenQueue, Token, TokenReader, TokenSender};
 
         #[test]
@@ -336,7 +420,7 @@ mod tests {
         }
     }
 
-    mod streamed_token_channel {
+    mod parallel_token_queue {
         use super::{ParallelTokenQueue, Token, TokenReader, TokenSender};
 
         #[test]
@@ -409,14 +493,16 @@ mod tests {
         }
     }
 
-    mod generator_token_channel {
-        use super::{GeneratorTokenQueue, Token, TokenReader};
+    mod generator_token_queue {
+        use super::{
+            GeneratorTokenQueue, GeneratorTokenQueueBuffer, Token, TokenReader, TokenSender,
+        };
 
-        fn lexer(state: &mut u8) -> Option<Token<u8, ()>> {
+        fn lexer(state: &mut u8, sender: &mut GeneratorTokenQueueBuffer<u8, ()>) {
             *state += 1;
             match state {
-                1 | 2 | 3 => Some(Token(*state * 2, ())),
-                _ => None,
+                1 | 2 | 3 => sender.push(Token(*state * 2, ())),
+                _ => {}
             }
         }
 
